from math import fabs

from compas.geometry import Plane
from compas.geometry import Circle
from compas.geometry import Sphere
from compas.geometry import Cylinder
from compas.geometry import Point
from compas.geometry import Vector
from compas.geometry import scale_vector
from compas.geometry import normalize_vector
from compas.geometry import add_vectors
from compas.geometry import length_vector

try:
    from compas_view2.shapes import Arrow
except ImportError:
    pass

from jax_fdm.visualization.artists import FDNetworkArtist


__all__ = ["FDNetworkViewerArtist"]


class FDNetworkViewerArtist(FDNetworkArtist):
    """
    An artist that draws a force density network to a viewer.
    """
    default_opacity = 0.75
    arrow_bodywidth = 0.012
    arrow_headportion = 0.12
    arrow_headwidth = 0.04

    def __init__(self, network, viewer, *args, **kwargs):
        super().__init__(network, *args, **kwargs)
        # NOTE: this attribute has to be handled by a ViewerArtist
        self.viewer = viewer

        self.viewer_edges = {}
        self.viewer_nodes = {}
        self.viewer_loads = {}
        self.viewer_reactions = {}

    # ==========================================================================
    # Properties
    # ==========================================================================

    @property
    def objects(self):
        """
        Yield the next viewer object.
        """
        objects = {}

        objects.update(self.viewer_edges)
        objects.update(self.viewer_nodes)
        objects.update(self.viewer_loads)
        objects.update(self.viewer_reactions)

        for obj in objects.values():
            yield obj

    # ==========================================================================
    # Add
    # ==========================================================================

    def add(self):
        """
        Add the elements of the network to the viewer.
        """
        if self.collection_nodes:
            self.viewer_nodes = self.add_nodes()
        if self.collection_edges:
            self.viewer_edges = self.add_edges()
        if self.collection_reactions:
            self.viewer_reactions = self.add_reactions()
        if self.collection_loads:
            self.viewer_loads = self.add_loads()

    # ==========================================================================
    # Update
    # ==========================================================================

    def update(self):
        """
        Update the elements of the network drawn by this artist.
        """
        if self.viewer_nodes:
            self.update_nodes()
        if self.viewer_edges:
            self.update_edges()
        if self.viewer_reactions:
            self.update_reactions()
        if self.viewer_loads:
            self.update_loads()

    # ==========================================================================
    # Edges
    # ==========================================================================

    def draw_edge(self, edge, width, *args, **kwargs):
        """
        Draw an edge.
        """
        u, v = edge
        network = self.network

        plane = Plane(network.edge_midpoint(u, v), network.edge_direction(u, v))
        circle = Circle(plane, width / 2.0)

        return Cylinder(circle, height=network.edge_length(u, v))

    def add_edge(self, edge, color):
        """
        Add an edge to the viewer.
        """
        return self.viewer.add(edge,
                               facecolor=color,
                               linecolor=color,
                               show_edges=True,
                               opacity=self.default_opacity)

    def add_edges(self):
        """
        Add the edges of the network to the viewer.
        """
        edges = {}

        for edge, cylinder in self.collection_edges.items():
            color = self.edge_color[edge]
            obj = self.add_edge(cylinder, color)
            edges[edge] = obj

        return edges

    def update_edge(self, edge, width, color):
        """
        Update an edge.
        """
        u, v = edge
        network = self.network
        cylinder = self.collection_edges[edge]

        plane = Plane(network.edge_midpoint(u, v), network.edge_direction(u, v))
        circle = Circle(plane, width / 2)

        cylinder.circle = circle
        cylinder.height = self.network.edge_length(u, v)

        obj = self.viewer_edges[edge]
        obj.linecolor = color
        obj.facecolor = color

    def update_edges(self):
        """
        Update the edges of the network.
        """
        self.edge_color = self._init_edgecolor
        self.edge_width = self._init_edgewidth

        for edge in self.edges:
            width = self.edge_width[edge]
            color = self.edge_color[edge]
            self.update_edge(edge, width, color)

    # ==========================================================================
    # Draw one element
    # ==========================================================================

    def draw_node(self, node, size, *args, **kwargs):
        """
        Draw a node.
        """
        return Sphere(self.network.node_coordinates(node), radius=size/2.0)

    def add_node(self, node, color):
        """
        Add one node to the viewer.
        """
        return self.viewer.add(node,
                               facecolor=color,
                               linecolor=color,
                               show_edges=False,
                               opacity=self.default_opacity)

    def add_nodes(self):
        """
        Add the nodes to viewer.
        """
        nodes = {}

        for node, sphere in self.collection_nodes.items():
            color = self.node_color[node]
            obj = self.add_node(sphere, color)
            nodes[node] = obj

        return nodes

    def update_node(self, node, size):
        """
        Update a node in the viewer.
        """
        sphere = self.collection_nodes[node]
        sphere.point = self.network.node_coordinates(node)
        sphere.radius = size / 2.0

    def update_nodes(self):
        """
        Update the nodes in the viewer.
        """
        for node in self.collection_nodes.keys():
            size = self.node_size[node]
            self.update_node(node, size)

    # ==========================================================================
    # Loads
    # ==========================================================================

    def draw_load(self, node, scale, *args, **kwargs):
        """
        Draw a load vector at a node.
        """
        vector = self.network.node_load(node)

        if length_vector(vector) < self.load_tol:
            return

        xyz = self.network.node_coordinates(node)

        # shift start to make arrow head touch the node the load is applied to
        start = add_vectors(xyz, scale_vector(vector, -scale))

        # shift start to account for half size of edge thickness
        widths = []
        for edge in self.network.connected_edges(node):
            width = self.edge_width.get(edge)
            if not width:
                width = 0.0
            widths.append(width)

        start = add_vectors(start, scale_vector(normalize_vector(vector), -max(widths)))

        return self.draw_vector(vector, start, scale)

    def add_load(self, load, color):
        """
        Add one load to the viewer.
        """
        return self.viewer.add(load,
                               facecolor=color,
                               linecolor=color,
                               show_edges=True,
                               opacity=self.default_opacity)

    def add_loads(self):
        """
        Add the loads to viewer.
        """
        loads = {}

        for node, arrow in self.collection_loads.items():
            obj = self.add_load(arrow, self.default_loadcolor)
            loads[node] = obj

        return loads

    def update_load(self, node, scale):
        """
        Update a load in the viewer.
        """
        arrow_new = self.draw_load(node, scale)

        arrow = self.collection_loads[node]
        arrow.position = arrow_new.position
        arrow.direction = arrow_new.direction

    def update_loads(self):
        """
        Update the loads in the viewer.
        """
        for node in self.collection_loads.keys():
            self.update_load(node, self.load_scale)

    # ==========================================================================
    # Reaction forces
    # ==========================================================================

    def draw_reaction(self, node, scale, *args, **kwargs):
        """
        Draw a reaction vector at a node.
        """
        network = self.network

        vector = network.node_reaction(node)
        start = network.node_coordinates(node)

        if length_vector(vector) < self.reaction_tol:
            return

        # shift starting point if max force of connected edges is compressive
        connected_edges = network.connected_edges(node)
        if len(connected_edges) == 0:
            return

        forces = [network.edge_force(e) for e in connected_edges]
        max_force = max(forces, key=lambda f: fabs(f))
        if max_force < 0.0:
            start = add_vectors(start, scale_vector(vector, scale))

        # reverse vector to display direction of reaction forces
        return self.draw_vector(scale_vector(vector, -1.0), start, scale)

    def add_reaction(self, reaction, color):
        """
        Add one reaction force to the viewer.
        """
        return self.viewer.add(reaction,
                               facecolor=color,
                               linecolor=color,
                               show_edges=True,
                               opacity=self.default_opacity)

    def add_reactions(self):
        """
        Add the reaction forces to viewer.
        """
        reactions = {}

        for node, arrow in self.collection_reactions.items():
            color = self.reaction_color
            if isinstance(self.reaction_color, dict):
                color = color[node]
            obj = self.add_reaction(arrow, color)
            reactions[node] = obj

        return reactions

    def update_reaction(self, node, scale):
        """
        Update a reaction force in the viewer.
        """
        arrow_new = self.draw_reaction(node, scale)

        arrow = self.collection_reactions[node]
        arrow.position = arrow_new.position
        arrow.direction = arrow_new.direction

    def update_reactions(self):
        """
        Update the reaction forces in the viewer.
        """
        for node in self.collection_reactions.keys():
            self.update_reaction(node, self.reaction_scale)

    # ==========================================================================
    # Helpers
    # ==========================================================================

    def draw_vector(self, vector, start, scale):
        """
        Draw a vector.
        """
        vector_scaled = scale_vector(vector, scale)
        end = add_vectors(start, vector_scaled)

        return Arrow(Point(*start),
                     Vector.from_start_end(start, end),
                     head_portion=self.arrow_headportion,
                     head_width=self.arrow_headwidth,
                     body_width=self.arrow_bodywidth)



from functools import partial

import jax

import jax.numpy as jnp

from jax import vjp
from jax import custom_vjp
from jax import jacrev

from equinox.internal import while_loop

from jaxopt import FixedPointIteration
from jaxopt import AndersonAcceleration


# ==========================================================================
# Iterative solvers
# ==========================================================================

def solver_anderson(f, a, x_init, solver_config):
    """
    Solve for a fixed point of a function f(a, x) using anderson acceleration in jaxopt.
    """
    tmax = solver_config["tmax"]
    eta = solver_config["eta"]
    verbose = solver_config["verbose"]
    implicit_diff = solver_config["implicit"]

    def f_swapped(x, a):
        return f(a, x)

    fpi = AndersonAcceleration(fixed_point_fun=f_swapped,
                               maxiter=tmax,
                               tol=eta,  # 1e-5 is the jaxopt default
                               has_aux=False,
                               history_size=5,  # 5 is the jaxopt default
                               ridge=1e-6,  # 1e-5 is the jaxopt default
                               implicit_diff=implicit_diff,
                               verbose=verbose)

    result = fpi.run(x_init, a)

    return result.params


def solver_fixedpoint(f, a, x_init, solver_config):
    """
    Solve for a fixed point of a function f(a, x) using forward iteration in jaxopt.
    """
    tmax = solver_config["tmax"]
    eta = solver_config["eta"]
    verbose = solver_config["verbose"]

    def f_swapped(x, a):
        return f(a, x)

    fpi = FixedPointIteration(fixed_point_fun=f_swapped,
                              maxiter=tmax,
                              tol=eta,
                              has_aux=False,
                              # implicit_diff=True,
                              # jit=True,
                              # unroll=False,
                              verbose=verbose)

    result = fpi.run(x_init, a)

    return result.params


def solver_forward(f, a, x_init, solver_config):
    """
    Solve for a fixed point of a function f(a, x) using forward iteration.
    """
    tmax = solver_config["tmax"]
    eta = solver_config["eta"]
    verbose = solver_config["verbose"]

    def distance(x_prev, x):
        residual = jnp.mean(jnp.linalg.norm(x_prev - x, axis=1))
        if verbose:
            jax.debug.print("Residual: {}", residual)
        return residual

    def cond_fun(carry):
        x_prev, x = carry
        return distance(x_prev, x) > eta

    def body_fun(carry):
        _, x = carry
        return x, f(a, x)

    init_val = (x_init, f(a, x_init))

    _, x_star = while_loop(cond_fun,
                           body_fun,
                           init_val,
                           max_steps=tmax,
                           kind="checkpointed")

    return x_star


def solver_newton(f, a, x_init, solver_config):
    """
    Find a root of the equation f(a, x) - x = 0 using Newton's method.
    """
    def f_root(x):
        return f(a, x) - x

    u, v = x_init.shape
    nrows = ncols = u * v

    def f_newton(a, x):
        b = f_root(x)
        jinv = jacrev(f_root)(x)
        jinv = jnp.reshape(jinv, (nrows, ncols))
        b = jnp.reshape(b, (-1, 1))
        step = jnp.linalg.solve(jinv, b)
        step = jnp.reshape(step, (-1, 3))

        return x - step

    x_star = solver_forward(f_newton, a, x_init, solver_config)

    return jnp.reshape(x_star, (-1, 3))


# ==========================================================================
# Fixed point solver function
# ==========================================================================

@partial(custom_vjp, nondiff_argnums=(0, 1, 2))
def fixed_point(solver, solver_config, f, a, x_init):
    """
    Solve for a fixed point of a function f(a, x) using forward iteration.
    """
    return solver(f, a, x_init, solver_config)


def fixed_point_fwd(solver, solver_config, f, a, x_init):
    """
    The forward pass of a fixed point solver.
    """
    x_star = fixed_point(solver, solver_config, f, a, x_init)
    return x_star, (a, x_star)


def fixed_point_bwd(solver, solver_config, fn, res, x_star_bar):
    """
    The backward pass of a fixed point solver.
    """
    a, x_star = res
    _, vjp_a = vjp(lambda a: fn(a, x_star), a)

    def rev_iter(packed, u):
        a, x_star, x_star_bar = packed
        _, vjp_x = vjp(lambda x: fn(a, x), x_star)
        return x_star_bar + vjp_x(u)[0]

    partial_func = solver(rev_iter,
                          (a, x_star, x_star_bar),
                          x_star_bar,
                          solver_config)

    a_bar = vjp_a(partial_func)[0]

    return a_bar, None


fixed_point.defvjp(fixed_point_fwd, fixed_point_bwd)



from compas.geometry import Vector
from compas.geometry import Cone
from compas.geometry import Circle
from compas.geometry import Cylinder
from compas.geometry import Plane
from compas.geometry import Shape

from compas.datastructures import Mesh


__all__ = ["Arrow"]


class Arrow(Shape):
    """
    An arrow is defined by the position where it starts and direction vector where it points to.

    Notes
    -----
    This object is a verbatim copy of compas_view2.shapes.
    I paste it here because it is not available in COMPAS core, and
    we want to be able to visualize arrows in a notebook without having to
    install compas_view2.
    """
    def __init__(self,
                 position=[0, 0, 0],
                 direction=[0, 0, 1],
                 head_portion=0.3,
                 head_width=0.1,
                 body_width=0.02):
        super().__init__()
        self.position = Vector(*position)
        self.direction = Vector(*direction)
        self.head_portion = head_portion
        self.head_width = head_width
        self.body_width = body_width

    @property
    def data(self):
        """Returns the data dictionary that represents the Arrow.
        Returns
        -------
        dict
            The arrow data.
        """
        return {"position": list(self.position), "direction": list(self.direction)}

    @data.setter
    def data(self, data):
        self.position = Vector(*data["position"])
        self.direction = Vector(*data["direction"])

    # ==========================================================================
    # customization
    # ==========================================================================

    def __repr__(self):
        return "Arrow({0}, {1})".format(self.position, self.direction)

    # ==========================================================================
    # constructors
    # ==========================================================================

    @classmethod
    def from_data(cls, data):
        """Construct a Arrow from its data representation.

        Parameters
        ----------
        data : :obj:`dict`
            The data dictionary.
        Returns
        -------
        Arrow
            The constructed arrow.
        """
        arrow = cls()
        arrow.data = data
        return arrow

    # ==========================================================================
    # methods
    # ==========================================================================

    def to_vertices_and_faces(self, u=4):
        """Returns a list of vertices and faces.

        Parameters
        ----------
        u : int, optional
            Number of faces in the "u" direction.
            Default is ``4``.
        Returns
        -------
        (vertices, faces)
            A list of vertex locations and a list of faces,
            with each face defined as a list of indices into the list of vertices.
        """
        if u < 3:
            raise ValueError("The value for u should be u > 3.")

        # Head of arrow
        head_position = self.position + self.direction * (1 - self.head_portion)
        plane = Plane(head_position, self.direction)
        circle = Circle(plane, self.head_width * self.direction.length)
        cone = Cone(circle, self.direction.length * self.head_portion)
        v, f = cone.to_vertices_and_faces(u=u)
        head = Mesh.from_vertices_and_faces(v, f)

        # Body of arrow
        body_center = self.position + self.direction * (1 - self.head_portion) / 2
        plane = Plane(body_center, self.direction)
        circle = Circle(plane, self.body_width * self.direction.length)
        cylinder = Cylinder(circle, self.direction.length * (1 - self.head_portion))
        v, f = cylinder.to_vertices_and_faces(u=u)
        body = Mesh.from_vertices_and_faces(v, f)

        body.join(head)

        return body.to_vertices_and_faces()

    def transform(self, transformation):
        """Transform the Arrow.

        Parameters
        ----------
        transformation : :class:`Transformation`
            The transformation used to transform the cone.
        """
        self.position.transform(transformation)
        self.direction.transform(transformation)



import jax.numpy as jnp

from jax import vmap

from jax_fdm.goals import ScalarGoal

from jax_fdm.goals.mesh import MeshGoal
from jax_fdm.goals.network import NetworkXYZLaplacianGoal


class MeshXYZLaplacianGoal(NetworkXYZLaplacianGoal):
    """
    A thin wrapper of NetworkXYZLaplacianGoal.
    """
    def __init__(self, target=0.0, weight=1.0):
        super().__init__(key=-1, target=target, weight=weight)


class MeshXYZFaceLaplacianGoal(ScalarGoal, MeshGoal):
    """
    Minimize the Laplacian energy of the XYZ faces coordinates of a mesh.

    Notes
    -----
    This goal can be handy to "smoothen" the looks of a mesh.

    This energy is computed as the distance of the XYZ coordinates of
    every vertex to the centroid of the centroids of the faces the
    vertex is part of.

    An enegy-minimizing mesh will have every vertex as close as possible
    to its neighboring faces centroid.
    """
    def __init__(self, target=0.0, weight=1.0):
        super().__init__(key=-1, target=target, weight=weight)
        self.connectivity_faces_vertices = None

    def init(self, model, structure):
        """
        Initialize the goal with information of a model and a structure.
        """
        super().init(model, structure)
        self.connectivity_faces_vertices = structure.connectivity_faces_vertices

    def prediction(self, eq_state, index):
        """
        The current load path of the network.
        """
        def vertex_laplacian(vertex_xyz, vertex_mask, faces_centroid):
            nbrs_xyz = vertex_faces_centroid(vertex_mask, faces_centroid)
            return vertex_square_distance(vertex_xyz, nbrs_xyz)

        def vertex_faces_centroid(vertex_mask, faces_centroid):
            nbrs_xyzs = jnp.reshape(vertex_mask, (-1, 1)) * faces_centroid
            # return average centroid
            return jnp.sum(nbrs_xyzs, axis=0) / jnp.sum(vertex_mask)

        def vertex_square_distance(vertex_xyz, nbrs_xyz):
            return jnp.sum(jnp.square(vertex_xyz - nbrs_xyz))

        faces_centroid = self.connectivity_faces_vertices @ eq_state.xyz
        vertices_laplacian = vmap(vertex_laplacian, in_axes=(0, 1, None))
        laplacians = vertices_laplacian(eq_state.xyz,
                                        self.connectivity_faces_vertices,
                                        faces_centroid)

        laplacian = jnp.mean(laplacians)

        return jnp.atleast_1d(laplacian)



import numpy as np

from jax_fdm.geometry import closest_point_on_line

from jax_fdm.goals.node import NodePointGoal


class NodeLineGoal(NodePointGoal):
    """
    Pulls the xyz position of a node to a target line ray.
    """
    @property
    def target(self):
        """
        The target to achieve
        """
        return self._target

    @target.setter
    def target(self, target):
        self._target = np.reshape(target, (-1, 2, 3))

    @staticmethod
    def goal(target, prediction):
        """
        """
        return closest_point_on_line(prediction, target)


if __name__ == "__main__":

    import jax.numpy as jnp

    from jax import vmap

    from compas.geometry import Line

    points = [[0.0, 0.0, 0.0], [0.0, -0.5, 0.0], [0.0, 0.5, 0.0]]
    points = jnp.array(points)

    line = Line([1.0, 1.0, 0.0], [1.0, -1.0, 0.0])
    lines = [line] * len(points)
    lines = np.reshape(lines, (-1, 2, 3))

    closest = vmap(closest_point_on_line)(points, lines)
    print(closest)
    print(closest.shape)

    assert np.allclose(closest, jnp.array([[1.0, 0.0, 0.0], [1.0, -0.5, 0.0], [1.0, 0.5, 0.0]])), closest
    print("Done!")



"""
A gradient-based optimizer.
"""
from time import perf_counter
from itertools import groupby
from functools import partial

import jax.numpy as jnp

from jax import jit
from jax import grad
from jax import value_and_grad

from scipy.optimize import minimize
from scipy.optimize import Bounds

from jax_fdm.equilibrium import LoadState
from jax_fdm.equilibrium import EquilibriumParametersState

from jax_fdm.parameters import ParameterManager
from jax_fdm.parameters import EdgeForceDensityParameter

from jax_fdm.optimization import Collection


# ==========================================================================
# Optimizer
# ==========================================================================

class Optimizer:
    """
    Base class for all optimizers.
    """
    def __init__(self, name, disp=True):
        self.name = name
        self.disp = disp
        self.pm = None
        self.loads_static = None

    def constraints(self, constraints, model, params_opt):
        """
        Returns the defined constraints in a format amenable to `scipy.minimize`.
        """
        if constraints:
            print(f"Warning! {self.name} does not support constraints. I am ignoring them.")

    def gradient(self, loss):
        """
        Compute the gradient function of a loss function.
        """
        return jit(grad(loss, argnums=0))

    def hessian(self, loss):
        """
        Compute the hessian function of a loss function.
        """
        pass

# ==========================================================================
# Loss
# ==========================================================================

    def loss(self, params_opt, loss, model, structure):
        """
        The wrapper loss.
        """
        params = self.parameters_fdm(params_opt)

        return loss(params, model, structure)

# ==========================================================================
# Goals
# ==========================================================================

    def goals(self, loss, model, structure):
        """
        Pre-process the goals in the loss function to accelerate computations.
        """
        for term in loss.terms_error:
            goal_collections = self.collect_goals(term.goals)
            for goal_collection in goal_collections:
                goal_collection.init(model, structure)
            term.collections = goal_collections

# ==========================================================================
# Minimization
# ==========================================================================

    def problem(self,
                model,
                structure,
                network,
                loss,
                parameters=None,
                constraints=None,
                maxiter=100,
                tol=1e-6,
                callback=None,
                jit_fn=True):
        """
        Set up an optimization problem.
        """
        # optimization parameters
        if not parameters:
            parameters = [EdgeForceDensityParameter(edge) for edge in network.edges()]

        self.pm = ParameterManager(model, parameters, structure, network)
        x = self.parameters_value()

        # message
        print(f"\n***Constrained form finding***\nParameters: {len(x)} \tGoals: {loss.number_of_goals()}")

        # parameter bounds
        bounds = self.parameters_bounds()

        assert x.size == self.pm.bounds_low.size
        assert x.size == self.pm.bounds_up.size

        # build goal collections
        self.goals(loss, model, structure)
        print(f"\tGoal collections: {loss.number_of_collections()}\n\tRegularizers: {loss.number_of_regularizers()}")

        # load matters
        loads = LoadState.from_datastructure(network)
        self.loads_static = loads.edges, loads.faces

        loss_fn = partial(self.loss, loss=loss, model=model, structure=structure)
        loss_and_grad_fn = value_and_grad(loss_fn)
        if jit_fn:
            loss_and_grad_fn = jit(loss_and_grad_fn)

        print("Warming up the pressure cooker...")
        start_time = perf_counter()
        loss_val, grad_val = loss_and_grad_fn(x)
        print(f"\tLoss and grad warmup time: {(perf_counter() - start_time):.4} seconds")
        print(f"\tInitial loss value: {loss_val:.4}")
        print(f"\tInitial gradient norm: {jnp.linalg.norm(grad_val):.4}")
        assert jnp.sum(jnp.isnan(grad_val)) == 0, "NaNs found in gradient calculation!"

        # gradient of the loss function
        hessian_fn = self.hessian(loss_fn)  # w.r.t. first function argument
        if hessian_fn:
            if jit_fn:
                hessian_fn = jit(hessian_fn)
            start_time = perf_counter()
            _ = hessian_fn(x)
            print(f"\tHessian warmup time: {(perf_counter() - start_time):.4} seconds")

        # constraints
        constraints = constraints or []
        if constraints:
            start_time = perf_counter()
            constraints = self.constraints(constraints, model, structure, x)
            print(f"\tConstraints warmup time: {(perf_counter() - start_time):.4} seconds")

        opt_kwargs = {"fun": loss_and_grad_fn,
                      "jac": True,
                      "hess": hessian_fn,
                      "method": self.name,
                      "x0": x,
                      "tol": tol,
                      "bounds": bounds,
                      "constraints": constraints,
                      "callback": callback,
                      "options": {"maxiter": maxiter, "disp": self.disp}}

        return opt_kwargs

    def solve(self, opt_problem):
        """
        Solve an optimization problem by minimizing a loss function via gradient descent.
        """
        print(f"Optimization with {self.name} started...")
        start_time = perf_counter()

        # minimize
        res_q = self._minimize(opt_problem)
        loss_and_grad_fn = opt_problem["fun"]
        loss_val, grad_val = loss_and_grad_fn(res_q.x)

        print(res_q.message)
        print(f"Final gradient norm: {jnp.linalg.norm(grad_val):.4}")
        print(f"Final loss in {res_q.nit} iterations: {res_q.fun:.4}")
        print(f"Optimization elapsed time: {perf_counter() - start_time} seconds")

        return res_q.x

    @staticmethod
    def _minimize(opt_problem):
        """
        Scipy backend method to minimize a loss function.
        """
        return minimize(**opt_problem)

# ==========================================================================
# Parameters
# ==========================================================================

    def parameters_bounds(self):
        """
        Return a tuple of arrays with the upper and the lower bounds of optimization parameters.
        """
        return Bounds(lb=self.pm.bounds_low, ub=self.pm.bounds_up)

    def parameters_value(self):
        """
        Return a flat array with the value of the optimization parameters.
        """
        return self.pm.parameters_value

    def parameters_fdm(self, params_opt):
        """
        Reconstruct the force density parameters from the optimization parameters.
        """
        params = self.pm.parameters_fdm(params_opt)

        q, xyz_fixed, loads_nodes = params
        loads_edges, loads_faces = self.loads_static

        loads = LoadState(nodes=loads_nodes,
                          edges=loads_edges,
                          faces=loads_faces)

        return EquilibriumParametersState(q=q,
                                          xyz_fixed=xyz_fixed,
                                          loads=loads)

# ==========================================================================
# Goal collections
# ==========================================================================

    @staticmethod
    def collect_goals(goals):
        """
        Convert a list of goals into a list of goal collections.
        """
        goals_collectable = []
        goals_uncollectable = []

        for goal in goals:
            if goal.is_collectible:
                goals_collectable.append(goal)
            else:
                goals_uncollectable.append(goal)

        collections = []

        if goals_collectable:
            goals_sorted = sorted(goals_collectable, key=lambda g: type(g).__name__)
            groups = groupby(goals_sorted, lambda g: type(g))

            for _, group in groups:
                group = list(group)
                collection = Collection(group)
                collections.append(collection)

        if goals_uncollectable:
            for goal in goals_uncollectable:
                collections.append(Collection([goal]))

        return collections



from functools import partial

import jax.numpy as jnp

from jax import jit
from jax import vjp
from jax import jacfwd

from jax_fdm.optimization.optimizers import SecondOrderOptimizer
from jax_fdm.optimization.optimizers import ConstrainedOptimizer

try:
    from cyipopt import minimize_ipopt
except (ImportError, ModuleNotFoundError):
    pass


# ==========================================================================
# Constrained optimizer
# ==========================================================================

class IPOPT(ConstrainedOptimizer, SecondOrderOptimizer):
    """
    Interior Point Optimizer (Ipopt) for large-scale, gradient-based nonlinear optimization

    The optimizer supports box bounds on the optimization parameters, and equality and inequality constraints.

    Notes
    -----
    Ipopt expects that the loss and constraint functions are twice differentiable.
    """
    def __init__(self, acc_tol=1e-9, disp=1, **kwargs):
        super().__init__(name="IPOPT", disp=disp, **kwargs)
        self.acceptable_tol = acc_tol

    def _minimize(self, opt_problem):
        """
        Cyipopt backend method to minimize a loss function.
        """
        opt_problem["options"]["acceptable_tol"] = self.acceptable_tol

        return minimize_ipopt(**opt_problem)

    def constraint_eq(self, params_opt, constraint, model):
        """
        A wrapper function for an equality constraint on the parameters.
        """
        return self.constraint_ineq_low(params_opt, constraint, model)

    def constraint_ineq_up(self, params_opt, constraint, model):
        """
        A wrapper function for an inequality constraint on an upper bound of the parameters.
        """
        return constraint.bound_up - self.constraint(params_opt, constraint, model)

    def constraint_ineq_low(self, params_opt, constraint, model):
        """
        A wrapper function for an inequality constraint on a lower bound of the parameters.
        """
        return self.constraint(params_opt, constraint, model) - constraint.bound_low

    @staticmethod
    def hvp(x, v, f):
        """
        Calculate the product of the second derivatives of the vector-valued constraint function and a vector of Lagrange multipliers.
        """
        def _vjp(s):
            _, vjp_fun = vjp(f, s)
            return vjp_fun(v)[0]

        return jacfwd(_vjp)(x)

    def parameters_bounds(self):
        """
        Return a tuple of arrays with the upper and the lower bounds of optimization parameters.
        """
        return list(zip(self.pm.bounds_low, self.pm.bounds_up))

    def constraints(self, constraints, model, params_opt):
        """
        Returns the defined constraints in a format amenable to `cyipopt`.
        """
        if not constraints:
            return []

        print(f"Constraints: {len(constraints)}")
        constraints = self.collect_constraints(constraints)
        print(f"\tConstraint colections: {len(constraints)}")

        clist = []
        for constraint in constraints:

            # initialize constraint
            constraint.init(model)

            # select constraint functions
            funs = []
            if jnp.allclose(constraint.bound_low, constraint.bound_up):
                ctype = "eq"
                funs.append(self.constraint_eq)
                print("\tAdding one equality constraint")
            else:
                ctype = "ineq"
                if not jnp.allclose(constraint.bound_low, -jnp.inf):
                    print("\tAdding bound low inequality constraint")
                    funs.append(self.constraint_ineq_low)
                if not jnp.allclose(constraint.bound_up, jnp.inf):
                    print("\tAdding bound high inequality constraint")
                    funs.append(self.constraint_ineq_up)
            if len(funs) == 0:
                print("\tNo constraints were processed. Check the constraint bounds!")
                return clist

            # format constraint functions in a dictionary
            for fun in funs:

                cdict = {}

                cfun = partial(fun, constraint=constraint, model=model)
                jac = jit(jacfwd(cfun))
                hvp = jit(partial(self.hvp, f=cfun))

                # warm start
                c = cfun(params_opt)
                _ = jac(params_opt)
                _ = hvp(params_opt, jnp.ones_like(c))

                # store
                cdict["type"] = ctype
                cdict["fun"] = cfun
                cdict["jac"] = jac
                cdict["hess"] = hvp

                clist.append(cdict)

        return clist



from jax import jacfwd
from jax import jacrev

from jax_fdm.optimization.optimizers import Optimizer


# ==========================================================================
# Second-order optimizer
# ==========================================================================

class SecondOrderOptimizer(Optimizer):
    """
    A gradient-based optimizer that uses the hessian to accelerate convergence.
    """
    def hessian(self, loss):
        """
        Compute the hessian function of a loss function.
        """
        # NOTE: jacrev(jacfwd) is x3 slower than hessian. Why?
        # NOTE: Ah, but jacfwd(jacrev) is as fast as hessian
        return jacfwd(jacrev(loss, argnums=0))



"""
A force density network.
"""

from compas.datastructures import Network

from jax_fdm.datastructures import FDDatastructure


class FDNetwork(Network, FDDatastructure):
    """
    A force density network.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.update_default_edge_attributes({"q": 0.0,
                                             "length": 0.0,
                                             "force": 0.0,
                                             "px": 0.0,
                                             "py": 0.0,
                                             "pz": 0.0})

        self.update_default_node_attributes({"x": 0.0,
                                             "y": 0.0,
                                             "z": 0.0,
                                             "px": 0.0,
                                             "py": 0.0,
                                             "pz": 0.0,
                                             "rx": 0.0,
                                             "ry": 0.0,
                                             "rz": 0.0,
                                             "is_support": False})

    # ----------------------------------------------------------------------
    # Constructors
    # ----------------------------------------------------------------------

    @classmethod
    def from_mesh(cls, mesh):
        """
        Create a force density network from a mesh.
        """
        nodes = {vkey: mesh.vertex_coordinates(vkey) for vkey in mesh.vertices()}
        network = cls.from_nodes_and_edges(nodes, mesh.edges())

        for node in network.nodes():
            attrs = mesh.vertex_attributes(node)
            network.node_attributes(node, names=attrs.keys(), values=attrs.values())

        for edge in network.edges():
            attrs = mesh.edge_attributes(edge)
            network.edge_attributes(edge, names=attrs.keys(), values=attrs.values())

        return network

    # ----------------------------------------------------------------------
    # Nodes
    # ----------------------------------------------------------------------

    def nodes_coordinates(self, keys=None, axes="xyz"):
        """
        Gets or sets the x, y, z coordinates of a list of nodes.
        """
        keys = keys or self.nodes()
        return [self.node_coordinates(node, axes) for node in keys]

    def nodes_fixedcoordinates(self, keys=None, axes="xyz"):
        """
        Gets the x, y, z coordinates of the anchors of the network.
        """
        if keys:
            keys = {key for key in keys if self.is_node_support(key)}
        else:
            keys = self.nodes_fixed()

        return [self.node_coordinates(node, axes) for node in keys]

    def number_of_anchors(self):
        """
        The number of anchored nodes.
        """
        return len(list(self.nodes_anchors()))

    def number_of_supports(self):
        """
        The number of supported nodes.
        """
        return len(list(self.nodes_supports()))

    def node_support(self, key):
        """
        Sets a node as a fixed anchor.
        """
        return self.node_attribute(key=key, name="is_support", value=True)

    def node_anchor(self, key):
        """
        Sets a node as a fixed anchor.
        """
        return self.node_support(key)

    def is_node_support(self, key):
        """
        Test if the node is a fixed node.
        """
        return self.node_attribute(key=key, name="is_support")

    def nodes_supports(self, keys=None):
        """
        Gets or sets the node keys where a support has been assigned.
        """
        if keys is None:
            return self.nodes_where({"is_support": True})

        return self.nodes_attribute(name="is_support", value=True, keys=keys)

    def nodes_fixed(self, keys=None):
        """
        Gets or sets the node keys where a support has been assigned.
        """
        return self.nodes_supports(keys)

    def nodes_anchors(self, keys=None):
        """
        Gets or sets the node keys where an anchor has been assigned.
        """
        return self.nodes_supports(keys)

    def nodes_free(self):
        """
        The keys of the nodes where there is no support assigned.
        """
        return self.nodes_where({"is_support": False})

    def node_load(self, key, load=None):
        """
        Gets or sets a load to the nodes of the network.
        """
        return self.node_attributes(key=key, names=("px", "py", "pz"), values=load)

    def nodes_loads(self, load=None, keys=None):
        """
        Gets or sets a load to the nodes of the network.
        """
        return self.nodes_attributes(names=("px", "py", "pz"), values=load, keys=keys)

    def nodes_residual(self, keys=None):
        """
        Gets the residual forces of the nodes of the network.
        """
        return self.nodes_attributes(names=("rx", "ry", "rz"), keys=keys)

    def node_residual(self, key):
        """
        Gets the residual force of a single node of the network.
        """
        return self.node_attributes(key=key, names=("rx", "ry", "rz"))

    def nodes_reactions(self, keys=None):
        """
        Gets the reaction forces of the nodes of the network.
        """
        keys = keys or self.nodes_fixed()
        return self.nodes_attributes(names=("rx", "ry", "rz"), keys=keys)

    def node_reaction(self, key):
        """
        Gets the reaction force of a single node of the network.
        """
        return self.node_attributes(key=key, names=("rx", "ry", "rz"))

    # ----------------------------------------------------------------------
    # Edges
    # ----------------------------------------------------------------------

    def is_edge_supported(self, key):
        """
        Test if any of the two nodes connected by the edge is a support.
        """
        return any(self.is_node_support(node) for node in key)

    def is_edge_fully_supported(self, key):
        """
        Test if the two nodes connected the edge are a support.
        """
        return all(self.is_node_support(node) for node in key)

    # ----------------------------------------------------------------------
    # Datastructure properties
    # ----------------------------------------------------------------------

    def parameters(self):
        """
        Return the design parameters of the network.
        """
        q = self.edges_forcedensities()
        xyz_fixed = self.nodes_fixedcoordinates()
        loads = self.nodes_loads()

        return q, xyz_fixed, loads



from time import perf_counter

import matplotlib.pyplot as plt

import numpy as np

from jax import vmap

import jax.numpy as jnp
import jax.tree_util as jtu

from jax_fdm import DTYPE_JAX

from jax_fdm.equilibrium import EquilibriumModel
from jax_fdm.equilibrium import structure_from_datastructure


class LossPlotter:
    """
    Plot a loss function.
    """
    def __init__(self, loss, datastructure, *args, **kwargs):
        self.loss = loss
        self.structure = structure_from_datastructure(datastructure, sparse=False)
        self.fig = plt.figure(**kwargs)

    def plot(self, history, print_breakdown=True, plot_legend=True):
        """
        Plot the loss function and its error components on a list of fdm parameter states.
        """
        print("\nPlotting loss function...")
        start_time = perf_counter()

        # Create batched parameter state
        params = jtu.tree_map(lambda leaf: jnp.asarray(leaf, dtype=DTYPE_JAX),
                              history,
                              is_leaf=lambda y: isinstance(y, list))

        # Model is dense because it dense supports vmapping and sparse does not
        model = EquilibriumModel(tmax=1)

        equilibrium_vmap = vmap(model, in_axes=(0, None))
        eq_states = equilibrium_vmap(params, self.structure)

        if print_breakdown:
            print("\n***Error breakdown***")

        errors_all = []

        for error_term in self.loss.terms_error:
            errors = vmap(error_term)(eq_states)
            errors_all.append(errors)
            plt.plot(errors, label=error_term.name)

            if print_breakdown:
                self.print_error_stats(error_term, errors)

        for reg_term in self.loss.terms_regularization:
            errors = vmap(reg_term)(params)
            errors_all.append(errors)
            plt.plot(errors, label=reg_term.name)

            if print_breakdown:
                self.print_error_stats(reg_term, errors)
                print()

        losses = jnp.sum(jnp.asarray(errors_all, dtype=DTYPE_JAX), axis=0)

        plt.plot(losses, label=self.loss.name)

        plt.xlabel("Optimization iterations")
        plt.ylabel("Loss")
        plt.yscale("log")
        plt.grid()

        if plot_legend:
            plt.legend()

        print(f"Plotting time: {(perf_counter() - start_time):.4} seconds")

        return losses

    def show(self):
        """
        Display the plot.
        """
        plt.show()

    @staticmethod
    def print_error_stats(error_term, errors):
        """
        Print error statistics
        """
        stats = {"first": errors[0],
                 "last": errors[-1],
                 "min": np.min(errors),
                 "max": np.max(errors)}

        name_string = "{:<18}\t".format(error_term.name)
        values_string = "  ".join(["{}: {:>12.4f}".format(key.capitalize(), value) for key, value in stats.items()])
        print(name_string + values_string)



import jax.numpy as jnp


# ==========================================================================
# Error
# ==========================================================================

class Error:
    """
    The base class for an error term in a loss function.
    """
    def __init__(self, goals, alpha=1.0, name=None, *args, **kwargs):
        self.goals = goals
        self.alpha = alpha
        self.name = name or self.__class__.__name__
        self.collections = []

    @staticmethod
    def error(gstate):
        raise NotImplementedError

    @staticmethod
    def errors(errors):
        return jnp.sum(errors)

    def __call__(self, eqstate):
        """
        Return the current value of the error term.
        """
        errors = []
        for goal_collection in self.collections:
            gstate = goal_collection(eqstate)
            error = self.error(gstate)
            errors.append(error)

        return self.errors(jnp.array(errors)) * self.alpha

    def number_of_goals(self):
        """
        The total number of individual goals in this error term.
        """
        return len(self.goals)

    def number_of_collections(self):
        """
        The total number of goal collections in this error term.
        """
        return len(self.collections)


# ==========================================================================
# Precooked error terms
# ==========================================================================

class SquaredError(Error):
    """
    The canonical squared error.

    It measures the L2 distance between the current and the target value of a goal.
    """
    @staticmethod
    def error(gstate):
        return jnp.sum(gstate.weight * jnp.square(gstate.prediction - gstate.goal))


class MeanSquaredError(SquaredError):
    """
    The seminal mean squared error.

    Average out all errors because no single error is important enough.
    """
    def errors(self, errors):
        return super(MeanSquaredError, self).errors(errors) / self.number_of_goals()


class RootMeanSquaredError(MeanSquaredError):
    """
    The root mean squared error.
    """
    def errors(self, errors):
        error = super(RootMeanSquaredError, self).errors(errors)
        return jnp.sqrt(error)


class PredictionError(Error):
    """
    The prediction error.

    You lose when you get too much of something.
    """
    @staticmethod
    def error(gstate):
        return gstate.prediction * gstate.weight


class AbsoluteError(Error):
    """
    The canonical absolute error.

    It measures the absolute difference between the current and the target value of a goal.
    """
    @staticmethod
    def error(gstate):
        return jnp.sum(gstate.weight * jnp.abs(gstate.prediction - gstate.goal))


class MeanAbsoluteError(AbsoluteError):
    """
    The canonical mean absolute error.
    """
    def errors(self, errors):
        return super(MeanAbsoluteError, self).errors(errors) / self.number_of_goals()


class LogMaxError(Error):
    """
    The log error for constraints with a target maximum value that should not be exceeded.

    Helpful to deal with soft barrier constraints with an upper bound.
    """
    @staticmethod
    def error(gstate):
        difference = gstate.prediction - gstate.goal
        # NOTE: shifting difference so that error is log(1) = 0.0 at least.
        error = jnp.log(jnp.where(difference >= 0.0, difference, 0.0) + 1.0)

        return jnp.sum(gstate.weight * error)



import numpy as np

from jax_fdm.constraints.node import NodeNormalAngleConstraint


class NodeTangentAngleConstraint(NodeNormalAngleConstraint):
    """
    Constraints the angle formed by the node tangent and a vector between a lower and an upper bound.

    The node tangent is calculated as 90º minus the node normal.
    """
    def __init__(self, key, vector, bound_low=None, bound_up=None):
        super().__init__(key=key, vector=vector, bound_low=bound_low, bound_up=bound_up)

    def constraint(self, eqstate, index):
        """
        Returns the curvature at a node based on the xyz coordinates of its one-hop neighborhood.
        """
        normal = super().constraint(eqstate, index)

        return np.pi * 0.5 - normal



from abc import abstractmethod
from abc import abstractproperty


# ==========================================================================
# Abstract goal
# ==========================================================================

class AbstractGoal:

    @abstractmethod
    def __call__(self, eqstate):
        """
        Return the current goal state.
        """
        raise NotImplementedError

    @abstractproperty
    def key(self):
        """
        The key of an element in a network.
        """
        raise NotImplementedError

    @abstractproperty
    def index(self):
        """
        The index of the goal key in the canonical ordering of a structure.
        """
        raise NotImplementedError

    @abstractproperty
    def weight(self):
        """
        The importance of the goal.
        """
        raise NotImplementedError

    @abstractmethod
    def prediction(self, eq_state):
        """
        The current reference value in the equilibrium state.
        """
        raise NotImplementedError

    @abstractmethod
    def target(self, prediction):
        """
        The target to achieve.
        """
        raise NotImplementedError



import numpy as np
import jax.numpy as jnp


def split_parameters(parray, func):
    """
    Split a flat array into flat subarrays given a filter function.
    """
    masks = func(parray)
    indices = [np.flatnonzero(mask) for mask in masks]
    sarrays = [parray[idx] for idx in indices]
    adef = np.argsort(np.concatenate(indices))

    return sarrays, adef


def combine_parameters(parrays, adef):
    """
    Combine a sequence of flat arrays given a filter function.
    """
    return jnp.concatenate(parrays)[adef]


def reshape_parameters(sarrays, shapes):
    """
    Reshape a sequence of flat arrays.
    """
    return (jnp.reshape(sarray, shape) for sarray, shape in zip(sarrays, shapes))



import numpy as np

from jax_fdm.goals.node import NodeNormalAngleGoal


class NodeTangentAngleGoal(NodeNormalAngleGoal):
    """
    Reach a target value for the angle formed by the node tangent and a reference vector.

    The node tangent is calculated as 90º minus the node normal.
    """
    def __init__(self, key, vector, target, weight=1.0):
        super().__init__(key=key, vector=vector, target=target, weight=weight)

    def prediction(self, eqstate, index):
        """
        Returns the angle between the node tangent and the reference vector.
        """
        angle_normal = super().prediction(eqstate, index)

        # TODO: take absolute of value? import jnp?
        angle_tangent = np.pi * 0.5 - angle_normal

        return angle_tangent



import jax.numpy as jnp

from jax_fdm.constraints.network import NetworkConstraint


class NetworkEdgesLengthConstraint(NetworkConstraint):
    """
    Set constraint bounds to the length of all the edges of a network.
    """
    @staticmethod
    def constraint(eqstate, model):
        """
        The constraint function relative to a equilibrium state.
        """
        return jnp.ravel(eqstate.lengths)



from .network import *  # noqa F403
from .loadpath import *  # noqa F403
from .laplacian import *  # noqa F403
from .smoothing import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



from compas_plotters import Plotter


class Plotter(Plotter):
    """
    A thin wrapper on the :class:`compas_plotters.plotter.Plotter`.

    This object exists for API consistency.
    """
    def __init__(self, *args, **kwargs):
        super(Plotter, self).__init__(*args, **kwargs)



import numpy as np
from scipy.sparse import coo_matrix

import jax
import jax.numpy as jnp

import equinox as eqx

from compas.numerical import connectivity_matrix
from compas.numerical import face_matrix as compas_face_matrix
from compas.utilities import pairwise

from jax.experimental.sparse import BCOO

from jax_fdm import DTYPE_NP
from jax_fdm import DTYPE_INT_NP
from jax_fdm import DTYPE_JAX

from jax_fdm.equilibrium.structures.mixins import IndexingMixins
from jax_fdm.equilibrium.structures.mixins import MeshIndexingMixins


# ==========================================================================
# Graphs
# ==========================================================================

class Graph(eqx.Module, IndexingMixins):
    """
    A graph.
    """
    nodes: np.ndarray
    edges: np.ndarray
    edges_indexed: jax.Array
    connectivity: jax.Array
    adjacency: jax.Array

    def __init__(self, nodes, edges):
        self.nodes = nodes

        assert edges.shape[1] == 2, "Edges in graph must connect exactly 2 nodes"
        self.edges = edges
        self.edges_indexed = self._edges_indexed()

        self.connectivity = self._connectivity_matrix()
        self.adjacency = self._adjacency_matrix()

    @property
    def num_nodes(self):
        """
        The number of nodes.
        """
        return self.nodes.size

    @property
    def num_edges(self):
        """
        The number of edges.
        """
        return self.edges.shape[0]

    def _connectivity_matrix(self):
        """
        The connectivity matrix between edges and nodes.
        """
        edges_indexed = self.edges_indexed

        return jnp.array(connectivity_matrix(edges_indexed, "array"), dtype=DTYPE_JAX)

    def _adjacency_matrix(self):
        """
        The adjacency matrix between nodes and nodes.
        """
        edges_indexed = self.edges_indexed

        return jnp.array(adjacency_matrix(edges_indexed, "array"), dtype=DTYPE_JAX)


class GraphSparse(Graph):
    """
    A sparse graph.
    """
    def _connectivity_matrix(self):
        """
        The connectivity matrix between edges and nodes in JAX format.

        Notes
        -----
        This currently is a dense array, but it should be a sparse one.

        How come?

        Currently there is a JAX bug that prevents us from using the
        sparse format with the connectivity matrix:

          C =  BCOO.from_scipy_sparse(self.connectivity_scipy)

        When not using a dense array from the next line, we get the
        following error:

          TypeError: float() argument must be a string or a number, not 'Zero'

        Therefore, we use the connectivity matrix method from the parent
        class, which outputs a dense array.

        However, submatrices connectivity_free and connectivity_fixed
        are correctly initialized and used as sparse matrices.
        """
        # C = super()._connectivity_matrix()
        # return BCOO.fromdense(C).astype(DTYPE_JAX)

        # C = self.connectivity_scipy
        # return BCOO.from_scipy_sparse(C)[:, :]

        # C = self.connectivity_scipy
        # args = (C.data, C.indices, C.indptr)
        # return CSC(args, shape=C.shape)

        return super()._connectivity_matrix()

    @property
    def connectivity_scipy(self):
        """
        The connectivity matrix between edges and nodes in SciPy CSC format.
        """
        # TODO: Refactor GraphSparse to return a JAX sparse matrix instead
        edges_indexed = self.edges_indexed

        return connectivity_matrix(edges_indexed, "csc")

    def _adjacency_matrix(self):
        """
        The adjacency matrix between nodes and nodes.
        """
        edges_indexed = self.edges_indexed

        A = adjacency_matrix(edges_indexed, "coo")

        return BCOO.from_scipy_sparse(A).todense()


# ==========================================================================
# Mesh
# ==========================================================================

class Mesh(Graph, MeshIndexingMixins):
    """
    A mesh.
    """
    # The faces array can have rows of different lengths. How to handle it?
    # Using a tuple instead of an array?
    vertices: np.ndarray
    faces: np.ndarray
    faces_indexed: jax.Array
    connectivity_faces_vertices: jax.Array
    connectivity_edges_faces: jax.Array

    def __init__(self, vertices, faces, edges=None, **kwargs):

        if edges is None:
            edges = self._edges_from_faces(faces)

        self.vertices = vertices
        assert faces.shape[1] > 2, "Mesh faces must connect at least 3 vertices"

        self.faces = faces
        self.faces_indexed = self._faces_indexed()
        self.connectivity_faces_vertices = self._connectivity_faces_matrix()

        super().__init__(vertices, edges)

        self.connectivity_edges_faces = self._connectivity_edges_faces_matrix()

    @property
    def num_vertices(self):
        """
        The number of vertices.
        """
        return self.vertices.size

    @property
    def num_faces(self):
        """
        The number of faces.
        """
        return self.faces.shape[0]

    def _connectivity_edges_faces_matrix(self):
        """
        The connectivity matrix between edges and faces of a mesh.
        """
        C = np.zeros((self.num_edges, self.num_faces))

        for eindex, findex in enumerate(self._edges_faces()):
            C[eindex, findex] = 1.

        return jnp.asarray(C)

    def _edges_faces(self):
        """
        The connectivity matrix of the edges and the faces of a mesh.
        """
        edges_faces = []
        for u, v in self.edges:

            edge = (int(u), int(v))
            findices = []

            for findex, face in enumerate(self.faces):
                face = [vkey for vkey in face if vkey >= 0]
                face_loop = np.concatenate((face, face[:1]))
                for u, v in pairwise(face_loop):
                    # iterate one time up, one time clockwise, another counter-clockwise
                    halfedge1 = (int(u), int(v))
                    halfedge2 = (int(v), int(u))

                    if edge == halfedge1 or edge == halfedge2:
                        findices.append(findex)

            assert len(findices) <= 2

            edges_faces.append(tuple(findices))

        return tuple(edges_faces)

    @staticmethod
    def _edges_from_faces(faces):
        """
        The the edges of the mesh.

        Edges have no topological meaning on a mesh and are used only to
        store data.

        The edges are calculated by first looking at all the halfedges of the
        faces of the mesh, and then only storing the unique halfedges.
        """
        halfedges = []
        for face in faces:
            face_loop = np.concatenate((face, face[:1]))
            for u, v in pairwise(face_loop):
                halfedge = (int(u), int(v))
                halfedges.append(halfedge)

        edges = []
        visited = set()
        for u, v in halfedges:
            if (u, v) in visited or (v, u) in visited:
                continue
            edge = (u, v)
            visited.add(edge)
            edges.append(edge)

        return np.asarray(edges, dtype=DTYPE_INT_NP)

    def _connectivity_faces_matrix(self):
        """
        The connectivity matrix between faces and nodes.
        """
        F = face_matrix(self.faces_indexed, "array", normalize=True)

        return jnp.asarray(F)


class MeshSparse(Mesh, GraphSparse):
    """
    A sparse mesh.

    Notes
    -----
    The connectivity matrices are currently dense matrices instead of sparse
    (as they should be) because I have encountered issues with the gradient
    computation via backpropagation if I leave them as sparse matrices.

    The error I see is the following:
    "TypeError: float() argument must be a string or a number, not 'Zero'"

    Which presumably arises upon calling bcoo._bcoo_dot_general_transpose.
    """
    def _connectivity_faces_matrix(self):
        """
        The connectivity matrix between faces and nodes in sparse format.
        """
        F = face_matrix(self.faces_indexed, "csc", normalize=True)

        return BCOO.from_scipy_sparse(F).todense()

    def _connectivity_edges_faces_matrix(self):
        """
        The connectivity matrix between edges and faces of a mesh in sparse format.
        """
        C = np.zeros((self.num_edges, self.num_faces))

        for eindex, findex in enumerate(self._edges_faces()):
            C[eindex, findex] = 1.

        return jnp.asarray(C)


# ==========================================================================
# Main
# ==========================================================================

def mesh_edges_faces(mesh):
    """
    The connectivity matrix of the edges and the faces of a mesh.
    """
    face_index = {fkey: idx for idx, fkey in enumerate(mesh.faces())}

    edges_faces = []
    for u, v in mesh.edges():

        findices = []
        for fkey in mesh.edge_faces(u, v):

            if fkey is None:
                continue

            findex = face_index[fkey]
            findices.append(findex)

        assert len(findices) <= 2

        edges_faces.append(tuple(findices))

    return edges_faces


def mesh_connectivity_edges_faces(mesh):
    """
    The connectivity matrix between edges and faces of a mesh.
    """
    num_edges = len(list(mesh.edges()))
    num_faces = mesh.number_of_faces()

    connectivity = np.zeros((num_edges, num_faces))

    edges_faces = mesh_edges_faces(mesh)

    for eindex, findex in enumerate(edges_faces):
        connectivity[eindex, findex] = 1.

    return connectivity


def face_matrix(face_vertices, rtype="array", normalize=False):
    """
    Creates a face-vertex adjacency matrix that skips -1 vertex entries.
    """
    face_vertices_clean = []
    for face in face_vertices:
        face_clean = [vertex for vertex in face if vertex >= 0]
        face_vertices_clean.append(face_clean)

    return compas_face_matrix(face_vertices_clean, rtype, normalize)


def adjacency_matrix(edges, rtype="array"):
    """
    Creates a vertex-vertex adjacency matrix.

    It expects that vertices / nodes are continuously indexed (no skips),
    and that edges are indexed from 0 to len(vertices) / len(nodes).
    """
    num_vertices = np.max(np.ravel(edges)) + 1

    # rows and columns indices for the COO format
    rows = np.hstack([edges[:, 0], edges[:, 1]])  # add edges in both directions for undirected graph
    cols = np.hstack([edges[:, 1], edges[:, 0]])

    # data to fill in (all 1s for the existence of edges)
    data = np.ones(len(rows), dtype=DTYPE_NP)

    # create the COO matrix
    A = coo_matrix(
        (data, (rows, cols)),
        shape=(num_vertices, num_vertices)
    )

    # convert to floating point matrix
    return _return_matrix(A.asfptype(), rtype)


def _return_matrix(M, rtype):
    if rtype == "list":
        return M.toarray().tolist()
    if rtype == "array":
        return M.toarray()
    if rtype == "csr":
        return M.tocsr()
    if rtype == "csc":
        return M.tocsc()
    if rtype == "coo":
        return M.tocoo()
    return M


# ==========================================================================
# Main
# ==========================================================================

if __name__ == "__main__":

    from compas.datastructures import Mesh as CMesh
    from compas.numerical import adjacency_matrix as adjacency_matrix_compas

    num_nodes = 5
    nodes = list(range(num_nodes))
    edges = [edge for edge in pairwise(nodes)]

    nodes = jnp.array(nodes, dtype=jnp.int64)
    edges = jnp.array(edges, dtype=jnp.int64)

    graph = Graph(nodes, edges)
    # supports = np.zeros_like(nodes)
    # supports[0] = 1
    # supports[-1] = 1
    # supports = jnp.asarray(supports)
    # structure = EquilibriumStructure(nodes, edges, supports)
    # assert structure.num_supports == 2
    # print(structure)
    # print(structure.supports)
    # print(structure.nodes_free)
    # print(structure.nodes_fixed)
    # print(structure.nodes_freefixed)

    # print(graph.nodes)
    # print(graph.edges)
    # print(graph.num_nodes)
    # print(graph.num_edges)
    # print(graph.connectivity_matrix)

    graph_sparse = GraphSparse(nodes, edges)

    # print(graph_sparse.nodes)
    # print(graph_sparse.edges)
    # print(graph_sparse.num_nodes)
    # print(graph_sparse.num_edges)

    assert jnp.allclose(graph_sparse.connectivity,
                        graph.connectivity)

    cmesh = CMesh.from_meshgrid(2.0, 2)
    print(cmesh)

    # test connectivity edges faces
    C = mesh_connectivity_edges_faces(cmesh)
    cmesh_faces = [cmesh.face_vertices(fkey) for fkey in cmesh.faces()]
    # cmesh_faces[0].append(-1)

    vertices_array = np.asarray(list(cmesh.vertices()))
    faces_array = np.asarray(cmesh_faces)
    edges_array = np.asarray(list(cmesh.edges()))
    mesh = MeshSparse(vertices_array, faces_array, edges_array)

    jnp.allclose(C, mesh.connectivity_edges_faces)

    # test adjacency matrix
    vertex_index = cmesh.vertex_index()
    adjacency = [[vertex_index[nbr] for nbr in cmesh.vertex_neighbors(vertex)] for vertex in cmesh.vertices()]
    A_c = adjacency_matrix_compas(adjacency, rtype="array")

    A = mesh.adjacency.todense()
    jnp.allclose(A, A_c)

    # def f(g):
    #     return jnp.sum(jnp.square(g.nodes - 1.0))

    # y = f(graph)
    # print(y)

    # from jax import jit
    # jf = jit(f)
    # z = jf(graph)
    # assert y == z
    # print(y, z)

    # gjf = jit(grad(f))
    # w = gjf(graph)
    # print("w", w)

    print("All good, cowboy!")



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import
from __future__ import print_function

import io
from os import path

from setuptools import setup
from setuptools.command.develop import develop
from setuptools.command.install import install


here = path.abspath(path.dirname(__file__))


def read(*names, **kwargs):
    return io.open(
        path.join(here, *names),
        encoding=kwargs.get("encoding", "utf8")
    ).read()


long_description = read("README.md")
requirements = read("requirements.txt").split("\n")
optional_requirements = {}

setup(
    name="jax_fdm",
    version="0.8.5",
    description="Auto-differentiable and hardware-accelerated force density method",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/arpastrana/jax_fdm",
    author="Rafael Pastrana",
    author_email="arpastrana@princeton.edu",
    license="MIT license",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    keywords=["jax", "automatic-differentiation", "architecture", "form-finding", "structural-design"],
    project_urls={},
    packages=["jax_fdm"],
    package_dir={"": "src"},
    package_data={},
    data_files=[],
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    python_requires=">=3.7",
    extras_require=optional_requirements,
    entry_points={
        "console_scripts": [],
    },
    ext_modules=[],
)



"""
A collection of scipy-powered, gradient-based optimizers.
"""
import numpy as np

from jax_fdm.optimization.optimizers import Optimizer
from jax_fdm.optimization.optimizers import ConstrainedOptimizer
from jax_fdm.optimization.optimizers import SecondOrderOptimizer


# ==========================================================================
# Optimizers
# ==========================================================================

class SLSQP(ConstrainedOptimizer):
    """
    The sequential least-squares programming optimizer.
    """
    def __init__(self, **kwargs):
        super().__init__(name="SLSQP", **kwargs)


class LBFGSB(Optimizer):
    """
    The limited-memory Boyd-Fletcher-Floyd-Shannon-Byrd (LBFGSB) optimizer.
    """
    def __init__(self, **kwargs):
        super().__init__(name="L-BFGS-B", disp=0, **kwargs)


class LBFGSBS(Optimizer):
    """
    The limited-memory Boyd-Fletcher-Floyd-Shannon-Byrd (LBFGSB) optimizer.

    This version of LBFGSB ensures compatibility of JAX gradients with scipy.
    However, it is slower than standard LBFGSB.
    """
    def __init__(self, **kwargs):
        super().__init__(name="L-BFGS-B", disp=0, **kwargs)

    def gradient(self, loss):
        """
        Compute the gradient function of a loss function.
        """
        jit_gfunc = super().gradient(loss)
        return lambda x: np.array(jit_gfunc(x))


class BFGS(Optimizer):
    """
    The Boyd-Fletcher-Floyd-Shannon optimizer.
    """
    def __init__(self, **kwargs):
        super().__init__(name="BFGS", **kwargs)


class NewtonCG(SecondOrderOptimizer):
    """
    The truncated Newton method. It uses a CG method to the compute the search direction.
    """
    def __init__(self, **kwargs):
        super().__init__(name="Newton-CG", **kwargs)


class TruncatedNewton(Optimizer):
    """
    Minimize a scalar function of one or more variables using a truncated Newton (TNC) algorithm.
    """
    def __init__(self, **kwargs):
        super().__init__(name="TNC", **kwargs)


class TrustRegionConstrained(ConstrainedOptimizer):
    """
    A trust-region algorithm for constrained optimization.
    """
    def __init__(self, **kwargs):
        super().__init__(name="trust-constr", **kwargs)


class TrustRegionKrylov(SecondOrderOptimizer):
    """
    A trust-region optimization algorithm.

    It uses a nearly exact trust-region algorithm that only requires
    matrix vector products with the hessian matrix.
    """
    def __init__(self, **kwargs):
        super().__init__(name="trust-krylov", **kwargs)


class TrustRegionNewton(SecondOrderOptimizer):
    """
    A Newton conjugate gradient trust-region algorithm.
    A trust-region algorithm for unconstrained optimization.
    """
    def __init__(self, **kwargs):
        super().__init__(name="trust-ncg", **kwargs)


class TrustRegionExact(SecondOrderOptimizer):
    """
    A nearly exact trust-region optimization algorithm.
    """
    def __init__(self, **kwargs):
        super().__init__(name="trust-exact", **kwargs)



from .optimizer import *  # noqa F403
from .constrained import *  # noqa F403
from .second_order import *  # noqa F403
from .scipy import *  # noqa F403
from .ipopt import *  # noqa F403


__all__ = [name for name in dir() if not name.startswith('_')]



import numpy as np

from compas.data import Data

import jax.tree_util as jtu

from jax_fdm.equilibrium import LoadState
from jax_fdm.equilibrium import EquilibriumParametersState


# ==========================================================================
# Recorder
# ==========================================================================

class OptimizationRecorder(Data):
    """A recorder that stores data during the optimization process.
    """
    def __init__(self, optimizer=None):
        self.optimizer = optimizer
        self.history = self._init_history()

    def _init_history(self):
        loads = LoadState(nodes=[], edges=[], faces=[])
        return EquilibriumParametersState(q=[], xyz_fixed=[], loads=loads)

    def __call__(self, xk, *args, **kwargs):
        if self.optimizer:
            xk = self.optimizer.parameters_fdm(xk)
        self.record(xk)

    def __len__(self):
        return len(self.history[0])

    def __getitem__(self, index):
        def index_from_leaf(leaf):
            return leaf[index]

        return jtu.tree_map(index_from_leaf,
                            self.history,
                            is_leaf=lambda x: isinstance(x, list))

    def record(self, parameters):
        def append_file(data, file):
            file.append(data)

        jtu.tree_map(append_file, parameters, self.history)

    @property
    def data(self):
        def leaf_to_list(leaf):
            return np.asarray(leaf, dtype=np.float64).tolist()

        data = {}
        history_params = jtu.tree_map(leaf_to_list,
                                      self.history,
                                      is_leaf=lambda x: isinstance(x, list))

        data["history"] = {key: val for key, val in history_params._asdict().items()}

        return data

    @data.setter
    def data(self, data):
        def leaf_to_array(leaf):
            return np.asarray(leaf, dtype=np.float64)

        # q, xyz_fixed, loads = data["history"]
        history = data["history"]
        nodes, edges, faces = history["loads"]

        loads = LoadState(nodes=nodes, edges=edges, faces=faces)
        history_params = EquilibriumParametersState(q=history["q"],
                                                    xyz_fixed=history["xyz_fixed"],
                                                    loads=loads)

        history_params = jtu.tree_map(leaf_to_array,
                                      history_params,
                                      is_leaf=lambda x: isinstance(x, list))

        self.history = history_params



from jax_fdm.constraints import Constraint


class EdgeConstraint(Constraint):
    """
    Base class for all constraints that pertain to an edge of a network.
    """
    def index_from_model(self, model, structure):
        """
        The index of the edge key in an equilibrium structure.
        """
        try:
            return structure.edge_index[self.key]
        except TypeError:
            return tuple([structure.edge_index[k] for k in self.key])



from compas_plotters.artists import VectorArtist


class FDVectorPlotterArtist(VectorArtist):
    """
    An alternative way to plot vectors as arrows.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def draw(self):
        """
        Draw the vector.

        Returns
        -------
        None

        """
        if self.draw_point:
            self._point_artist = self.plotter.add(self.point, edgecolor=self.color)

        length = self.vector.length
        min_width = 0.01
        width = length * 0.012 + min_width
        x, y = self.point[:2]
        dx, dy = (self.vector)[:2]
        self.zorder = 5000

        self._mpl_vector = self.plotter.axes.arrow(
            x,
            y,
            dx,
            dy,
            width=width,
            length_includes_head=True,
            head_width=length * 0.06 + min_width,
            head_length=length * 0.12 + min_width,
            color=self.color,
            zorder=self.zorder,
            alpha=1.0,
            lw=0.0,
        )



from jax_fdm.geometry import normalize_vector

from jax_fdm.goals import VectorGoal
from jax_fdm.goals.edge import EdgeGoal


class EdgeDirectionGoal(VectorGoal, EdgeGoal):
    """
    Make the direction of the edge of a network to be parallel to a target vector.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The edge vector in the network.
        """
        vector = eq_state.vectors[index, :]
        return normalize_vector(vector)

    @staticmethod
    def goal(target, prediction):
        """
        The target vector.
        """
        return normalize_vector(target)



from math import radians

from compas.artists import Artist

try:
    from compas_notebook.app import App as NotebookApp
except ImportError:
    class NotebookApp:
        pass

from jax_fdm.datastructures import FDNetwork


__all__ = ["NotebookViewer"]


class NotebookViewer(NotebookApp):
    """
    A thin wrapper on :class:`compas_notebook.app.App`.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.artists = []

    def add(self, data, **kwargs):
        """
        Add a COMPAS data object to the notebook viewer.

        It adds a viewer argument if the object is a :class:`jax_fdm.datastructures.FDNetwork`

        Parameters
        ----------
        data: :class:`compas.geometry.Primitive` | :class:`compas.geometry.Shape` | :class:`compas.datastructures.Datastructure`
            A COMPAS data object.
        **kwargs : dict, optional
            Additional visualization options.

        Returns
        -------
        view_data :class:`compas_view2.objects.Object`
            A visualization object.
        """
        if not isinstance(data, (FDNetwork)):
            return super(NotebookViewer, self).add(data, **kwargs)

        if kwargs.get("as_wireframe"):
            del kwargs["as_wireframe"]
            return super(NotebookViewer, self).add(data, **kwargs)

        artist = Artist(data, viewer=self, context="Notebook", **kwargs)
        self.artists.append(artist)
        artist.draw()
        artist.add()

    def show(self, *args, **kwargs):
        """
        Display the current scene.
        """
        self.set_camera(angles=[radians(89.), 0., 0.])
        return super(NotebookViewer, self).show(viewer="notebook", *args, **kwargs)



import numpy as np

import jax.numpy as jnp

from jax_fdm.geometry import angle_vectors

from jax_fdm.constraints.edge import EdgeConstraint


class EdgeAngleConstraint(EdgeConstraint):
    """
    Constraints the angle formed by an edge and a vector between a lower and an upper bound.
    """
    def __init__(self, key, vector, bound_low, bound_up):
        super().__init__(key=key, bound_low=bound_low, bound_up=bound_up)
        self._vector = None
        self.vector = vector

    @property
    def vector(self):
        """
        The vector to take the angle with.
        """
        return self._vector

    @vector.setter
    def vector(self, vector):
        self._vector = jnp.reshape(jnp.asarray(vector), (-1, 3))

    def vectors(self):
        """
        Create a matrix of vectors.
        """
        matrix = np.zeros((max(self.index) + 1, 3))
        for vec, idx in zip(self.vector, self.index):
            matrix[idx, :] = vec
        return matrix

    def init(self, model, structure):
        """
        Initialize the constraint with information from an equilibrium model.
        """
        super().init(model, structure)
        self.vector = self.vectors()

    def constraint(self, eqstate, index):
        """
        Returns the angle between an edge in an equilibrium state and a vector.
        """
        vector = eqstate.vectors[index, :]
        return angle_vectors(vector, self.vector[index, :])



import numpy as np

from jax_fdm.geometry import closest_point_on_plane

from jax_fdm.goals.node import NodePointGoal


class NodePlaneGoal(NodePointGoal):
    """
    Pulls the xyz position of a node to a target plane.
    """
    @property
    def target(self):
        """
        The target to achieve.
        """
        return self._target

    @target.setter
    def target(self, target):
        self._target = np.reshape(target, (-1, 2, 3))

    @staticmethod
    def goal(target, prediction):
        """
        Calculate the closest point on the target plane given the current node coordinates.
        """
        return closest_point_on_plane(prediction, target)



import jax.numpy as jnp

from jax_fdm.goals import ScalarGoal

from jax_fdm.goals.network import NetworkGoal


class NetworkLoadPathGoal(ScalarGoal, NetworkGoal):
    """
    Make the total load path of a network to reach a target magnitude.

    The load path of an edge is the absolute value of the product of the
    the force on the edge time its length.
    """
    @staticmethod
    def prediction(eq_state, index):
        """
        The current load path of the network.
        """
        load_path = jnp.sum(jnp.abs(jnp.multiply(eq_state.lengths, eq_state.forces)))

        return jnp.atleast_1d(load_path)
